# encoding: utf-8
"""
cerform.depth
=============

Routines for handling water depth data

:copyright: Copyright 2013 Ifremer / Cersat.
:license: Released under GPL v3 license, see :ref:`license`.

.. sectionauthor:: Jeff Piolle <jfpiolle@ifremer.fr>
.. codeauthor:: Jeff Piolle <jfpiolle@ifremer.fr>
"""
import numpy


def depth_fofonoff(pressure, lat):
    """Return the depth in meters from pressure in decibars using Saunders and
    Fofonoff's method.

    Deep_sea rs., 1976, 23, 109-111.
    Formula refitted for 1980 equation of state units

    Args:
        pressure (float or numpy array of float): pressure in decibars
        lat (float or numpy array of float): latitude in degrees

    Return:
        float: depth in meters

    Check value :
    Depth= 9712.653 M for P= 10000 Decibars, Latitude= 30 Deg

    Above for standard ocean: T= 0 Deg Celsius; S= 35 (PSS-78)
    """
    l_p = pressure
    l_lat = lat

    l_x = numpy.sin(l_lat / 57.29578)
    l_x = l_x * l_x
    l_gr = (9.780318 * (1.0 + (5.2788E-3 + 2.36E-5 * l_x) * l_x) + 1.092E-6 *
            l_p
            )
    l_depth = (((-1.82E-15 * l_p + 2.279E-10) * l_p - 2.2512E-5) *
               l_p + 9.72659) * l_p

    l_depth = l_depth / l_gr

    return l_depth

if __name__ == '__main__':
    print "depth test"
    pressure = 10000
    lat = 30.
    print "pressure: {} db".format(pressure)
    print "lat: {} deg".format(lat)
    print "depth: {} m. Should be 9712.653 m.".format(depth_fofonoff(pressure,
                                                                     lat))

