"""
:author: Antoine Grouazel
@purpose: give a single input/output interface to call coare3 or coare4 methods
creation: 31/08/2015
"""
from numpy import ma,ndarray
import logging
import os
from relative_humidity import relative_humidity_method_bentamy

INPUTS_NEEDED = {#it is better to put all the inputs possible
                     'coare3':['u','us','ts','t','Qs','Q','Rs','Rl','rain','zi','P','zu','zt','zq','lat','jcool','jwave','twave','hwave'],
                     'coare3.5':['u','t','rh','ts','P','Rs','Rl','zu','zt','zq','lat','zi','rain','cp','hwave','jcool'],
                     'coare4.0':['u','t','rh','P','ts','Rs','Rl','lat','zi','zu','zt','zq']
#                               ['u','zu','t','zt','rh','zq','P','ts','Rs','Rl','lat','zi']
                     }
INPUTS_FLOAT = {#list parameters that don't have to be reshaped
                'coare3':['zu','zt','zq','jcool','jwave','zi'],
                'coare3.5':['zu','zt','zq','jcool'],
                'coare4.0':['zu','zt','zq','jcool']
                }

INPUTS_CST = {#inputs that have to be constant
              'coare3':[],
              'coare3.5':['zu','zt','zq'],
              'coare4.0':[],
              }

def get_relative_humidity(algo,inputs_given):
    """
    complete inputs arguments using combination of other ones
    Args:
        algo (str):
        inputs_given (dict):
    """
    if algo in ['coare3.5','coare4']:
        if 'rh' not in inputs_given.keys() and 't' in inputs_given.keys() and 'Q' in inputs_given.keys():
            inputs_given['rh'] = relative_humidity_method_bentamy(inputs_given['t'],inputs_given['Q']/1000.0)
            logging.info('compute relative humidity from t and Q (mean val: %1.2f)',inputs_given['rh'].mean())
    return inputs_given


def default_input_values(algo,shape,param=None):
    """
    purpose: read the default values files
    Args:
        shape (tuple): give the dimension of input field latxlon (e.g 720x1440)
        param (str): name of the variable you want
    Return:
        x (list or dict): containing default values to be used in coare3/4
    """
    vals = {}
    pwd = os.path.dirname(__file__)
    fid = open(os.path.join(pwd,'list_default_values.txt'),'r')
    lines = fid.readlines()
    fid.close()
    for ll in lines:
        ll = ll.strip('\n')
        key = ll.split('=')[0]
        value = float(ll.split('=')[1])
#         if key in ['zu','zt','zq','jcool','jwave','zi']:
        
        if key in INPUTS_FLOAT[algo]:
            vals[key] = value
        elif key=='lat':
            vals[key] = value*ma.ones((shape[0],))
        else:
            vals[key] = value*ma.ones(shape)
    if param is None:
        x = vals
    else:
        if param in vals.keys():
            x = vals[param]
        else:
            raise Exception('%s is not defined.'%param)
    return x

def check_inputs(algo,inputs_given):
    """
    check that each input given is an input that is needed in the bulk method
    """
    list_input_needed = INPUTS_NEEDED[algo]
    for ii in inputs_given.keys():
        if ii not in list_input_needed:
            raise Exception('input %s is not a valid parameter name for %s algo \n you have to use a name in this list: %s '%(ii,algo,list_input_needed))

def complete_missing_args(algo,inputs_given,shape_wanted=None,param=None):
    """
    put default value when the given args are not fully-filled
    Args:
        algo (str): name of the algorithm used (ex: coare3.0)
        inputs_given (dic):
        shape_wanted (tuple): shape of the fields
        param (str): parameter you want to fill with default value (optional, default is all the algo parameters)
    Returns:
        inputs_complet (dic):
    """
    check_inputs(algo,inputs_given)
    inputs_complet = {}
    logging.debug('default_values| complete_missing_args | algo: %s %s',algo,type(algo))
    if param is None:
        list_input_needed = INPUTS_NEEDED[algo]
    else:
        list_input_needed = [param]
    
    if shape_wanted is None:
        #find the shape of the fields
        max_length = 0
        for ii in inputs_given.keys():
            if isinstance(inputs_given[ii],ndarray)==False and ii not in INPUTS_FLOAT[algo]:
                raise Exception('%s argument must be numpy masked array (even unique value) here: %s'%(ii,inputs_given[ii]))
            if isinstance(inputs_given[ii],ndarray):
                leni = len(inputs_given[ii].shape)
                if leni>max_length:
                        shape_wanted = inputs_given[ii].shape
                        max_length = leni
        logging.info('shape of input and output fields: %s',shape_wanted)
    if shape_wanted is None:#security
        shape_wanted = (1,)
    #fill missing args
    inputs_given = get_relative_humidity(algo,inputs_given)
    for input_needed in list_input_needed:
        logging.debug('input_needed: %s',input_needed)
        
        if input_needed in inputs_given.keys():
            if isinstance(inputs_given[input_needed],ndarray)==False:
                inputs_given[input_needed] = ma.array([inputs_given[input_needed]])
            inputs_complet[input_needed] = inputs_given[input_needed]
            logging.info('you have provided %s = %s',input_needed,inputs_complet[input_needed].shape)
            logging.debug('provided %s = %s',input_needed,inputs_complet[input_needed])
        else:
            inputs_complet[input_needed] = default_input_values(algo,shape_wanted,input_needed)
            logging.info(' %s is default value',input_needed)
            logging.debug(' %s is default value: %s',input_needed,inputs_complet[input_needed])
        if input_needed in INPUTS_CST[algo] and isinstance(inputs_complet[input_needed],float)==False and isinstance(inputs_complet[input_needed],int)==False:
            logging.debug('set %s to cst value',input_needed)
            inputs_complet[input_needed] = inputs_complet[input_needed].ravel()[0][0] #get the first element by default
    
    return inputs_complet